// This may look like C code, but it's really -*- C++ -*-
/*
 * Copyright (C) 2008 Emweb bv, Herent, Belgium.
 *
 * See the LICENSE file for terms of use.
 */
#include <Wt/Chart/WChartPalette.h>

namespace Wt {
  namespace Chart {

WChartPalette::~WChartPalette()
{ }

  }
}
